@@ -2,10 +2,43 @@
 #include "Python.h"
 #include "structmember.h"
 
+#ifndef Py_ssize_t
+#define Py_ssize_t ssize_t
+#endif
+
+/* Py_VISIT and Py_CLEAR are defined here to be compatible with Python 2.3 */
+
+/* Utility macro to help write tp_traverse functions.
+ * To use this macro, the tp_traverse function must name its arguments
+ * "visit" and "arg".  This is intended to keep tp_traverse functions
+ * looking as much alike as possible.
+ */
+#ifndef Py_VISIT
+#define Py_VISIT(op)							\
+        do { 								\
+                if (op) {						\
+                        int vret = visit((PyObject *)(op), arg);	\
+                        if (vret)					\
+                                return vret;				\
+                }							\
+        } while (0)
+#endif
+
+#ifndef Py_CLEAR
+#define Py_CLEAR(op)				\
+        do {                            	\
+                if (op) {			\
+                        PyObject *tmp = (PyObject *)(op);	\
+                        (op) = NULL;		\
+                        Py_DECREF(tmp);		\
+                }				\
+        } while (0)
+#endif
+
 /* partial type by Hye-Shik Chang <perky@FreeBSD.org>
    with adaptations by Raymond Hettinger <python@rcn.com>
    
-   foldr, foldl, id, the compose type and the flip type
+   map, filter, foldr, foldl, id, the compose type and the flip type
    by Collin Winter <collinw@gmail.com>
 
    Copyright (c) 2004-2006 Python Software Foundation.
@@ -249,44 +282,66 @@ static PyTypeObject partial_type = {
 static PyObject *
 functional_foldl(PyObject *self, PyObject *args)
 {
-	PyObject *func, *start, *seq, *iter, *result, *func_args, *item;
+	PyObject *func, *start, *seq, *iter, *result = NULL, *func_args, *item;
+    int decref_start = 1;
 	
 	if (!PyArg_UnpackTuple(args, "foldl", 3, 3, &func, &start, &seq)) {
 		return NULL;
 	}
-	Py_INCREF(start);
+    
+    if (!PyCallable_Check(func)) {
+		PyErr_SetString(PyExc_TypeError,
+			"argument 1 to foldl() must be callable");
+		return NULL;
+	}
 	
 	iter = PyObject_GetIter(seq);
 	if (iter == NULL) {
 		PyErr_SetString(PyExc_TypeError,
 			"argument 3 to foldl() must support iteration");
-		Py_DECREF(start);
 		return NULL;
 	}
 	
-	result = start;
-	
-	while ((item = PyIter_Next(iter))) {
-		func_args = PyTuple_New(2);
-		PyTuple_SET_ITEM(func_args, 0, result);
+    func_args = PyTuple_New(2);
+    if (NULL == func_args) {
+        Py_DECREF(iter);
+        return NULL;
+    }
+    
+    Py_INCREF(start);
+    while ((item = PyIter_Next(iter))) {
+        PyTuple_SET_ITEM(func_args, 0, start);
 		PyTuple_SET_ITEM(func_args, 1, item);
-		
 		result = PyObject_CallObject(func, func_args);
-		Py_DECREF(func_args);
-		if (NULL == result) {
-			goto Fail;
-		}
-	}
-	if (PyErr_Occurred()) {
-		Py_DECREF(start);
-		goto Fail;
-	}
-	goto Succeed;
+		PyTuple_SET_ITEM(func_args, 0, NULL);
+        PyTuple_SET_ITEM(func_args, 1, NULL);
+        Py_DECREF(start);
+        decref_start = 0;
+        Py_DECREF(item);
 
-Fail:
-	Py_CLEAR(result);
-Succeed:
-	Py_DECREF(iter);
+		if (NULL == result) {
+            Py_DECREF(iter);
+            Py_DECREF(func_args);
+            return NULL;
+		}
+        start = result;
+	}
+    if (decref_start) {
+        Py_DECREF(start);
+    }
+    Py_DECREF(func_args);
+    Py_DECREF(iter);
+    
+	if (PyErr_Occurred()) {
+		Py_XDECREF(result);
+        return NULL;
+	}
+    
+    /* This is here so if we skipped the iter loop, we return start */
+    if (NULL == result) {
+        Py_INCREF(start);
+        return start;
+    }
 	return result;
 }
 
@@ -307,7 +362,6 @@ foldr(PyObject *func, PyObject *base, PyObject *iter)
 {
 	PyObject *item, *func_args, *func_return, *recursive;
 	
-	/* new reference */
 	item = PyIter_Next(iter);
 	if (NULL == item) {
 		if (PyErr_Occurred()) {
@@ -322,13 +376,17 @@ foldr(PyObject *func, PyObject *base, PyObject *iter)
 		return NULL;
 	}
 	
-	/* new reference */
 	func_args = PyTuple_New(2);
-	/* both steal references */
 	PyTuple_SET_ITEM(func_args, 0, item);
 	PyTuple_SET_ITEM(func_args, 1, recursive);
-	
 	func_return = PyObject_CallObject(func, func_args);
+
+    if (NULL == func_return && recursive == base) {
+        /* This is so base doesn't get decrefed twice: once by
+           Py_DECREF(func_args) and once in functional_foldr
+           after we return NULL below */
+        PyTuple_SET_ITEM(func_args, 1, NULL);
+    }
 	Py_DECREF(func_args);
 	return func_return;
 }
@@ -338,21 +396,29 @@ functional_foldr(PyObject *self, PyObject *args)
 {
 	PyObject *func, *base, *seq, *iter, *result;
 
-	if (!PyArg_UnpackTuple(args, "foldr", 3, 3, &func, &base, &seq)) {
+	if (!PyArg_UnpackTuple(args, "foldr", 3, 3, &func, &base, &seq))
 		return NULL;
-	}
-	Py_INCREF(base);
+        
+    if (!PyCallable_Check(func)) {
+        PyErr_SetString(PyExc_TypeError,
+			"argument 1 to foldr() must be callable");
+		return NULL;
+    }
 	
 	iter = PyObject_GetIter(seq);
-	if (!iter) {
+	if (NULL == iter) {
 		PyErr_SetString(PyExc_TypeError,
 			"argument 3 to foldr() must support iteration");
-		Py_DECREF(base);
 		return NULL;
 	}
 	
+    Py_INCREF(base);
 	result = foldr(func, base, iter);
 	Py_DECREF(iter);
+    if (NULL == result) {
+        Py_DECREF(base);
+    }
+    
 	return result;
 }
 
@@ -589,14 +655,14 @@ compose_new(PyTypeObject *type, PyObject *args, PyObject *kw)
 	if (compo == NULL)
 		return NULL;
         
-    compo->unpack = PyObject_IsTrue(unpack) ? 1 : 0;
-    Py_DECREF(unpack);
+    /* Note to self: don't decref unpack */
+    compo->unpack = PyObject_IsTrue(unpack);
 
+    Py_INCREF(inner_func);
 	compo->inner_func = inner_func;
-	Py_INCREF(inner_func);
 	
-	compo->outer_func = outer_func;
 	Py_INCREF(outer_func);
+	compo->outer_func = outer_func;
 	
 	return (PyObject *)compo;
 }
@@ -606,21 +672,26 @@ compose_call(composeobject *compo, PyObject *args, PyObject *kw)
 {
 	PyObject *ret_val, *inner_ret_val, *inner_tuple;
 
-	inner_ret_val = PyObject_Call(compo->inner_func, args, kw);
+    inner_ret_val = PyObject_Call(compo->inner_func, args, kw);
 	if (NULL == inner_ret_val) {
 		return NULL;
 	}
 	
     if (compo->unpack) {
-        ret_val = PyObject_CallObject(compo->outer_func, inner_ret_val);
+        PyObject *arg_tuple = PySequence_Tuple(inner_ret_val);
         Py_DECREF(inner_ret_val);
+        if (NULL == arg_tuple)
+            return NULL;
+    
+        ret_val = PyObject_CallObject(compo->outer_func, arg_tuple);
+        Py_DECREF(arg_tuple);
     }
     else {
 	    inner_tuple = PyTuple_New(1);
-	    /* steals the reference to inner_ret_val */
 	    PyTuple_SET_ITEM(inner_tuple, 0, inner_ret_val);
-	
 	    ret_val = PyObject_CallObject(compo->outer_func, inner_tuple);
+
+        /* will decref inner_ret_val, too */
         Py_DECREF(inner_tuple);
     }    
         
@@ -832,17 +903,189 @@ static PyTypeObject flip_type = {
 	PyObject_GC_Del,		/* tp_free */
 };
 
+/* functional.map ***********************************************************/
+
+static PyObject *
+functional_map(PyObject *self, PyObject *args)
+{
+    PyObject *seq, *func, *accu, *it, *item, *arg, *result;
+    int status;
+    
+    if (!PyArg_UnpackTuple(args, "map", 2, 2, &func, &seq))
+		return NULL;
+        
+    if (!PyCallable_Check(func)) {
+		PyErr_SetString(PyExc_TypeError,
+				"map() argument 1 must be callable");
+        return NULL;
+	}
+    
+    it = PyObject_GetIter(seq);
+    if (NULL == it) {
+        PyErr_SetString(PyExc_TypeError,
+				"map() argument 2 must be iterable");
+        return NULL;
+    }
+    
+    accu = PyList_New(0);
+    if (NULL == accu) {
+        Py_DECREF(it);
+        return NULL;
+    }
+    
+    arg = PyTuple_New(1);
+	if (NULL == arg) {
+		Py_DECREF(it);
+        Py_DECREF(accu);
+        return NULL;
+    }
+    
+    while ((item = PyIter_Next(it))) {
+        PyTuple_SET_ITEM(arg, 0, item);
+        result = PyObject_CallObject(func, arg);
+        PyTuple_SET_ITEM(arg, 0, NULL);
+        Py_DECREF(item);
+        if (NULL == result)
+            goto fail;
+
+        status = PyList_Append(accu, result);
+        Py_DECREF(result);
+        if (status < 0)
+            goto fail;
+    }
+    
+    if (PyErr_Occurred()) {
+        Py_CLEAR(accu);
+    }
+    
+    Py_DECREF(arg);
+    Py_DECREF(it);
+    return accu;
+    
+fail:
+    Py_DECREF(it);
+    Py_DECREF(arg);
+    Py_DECREF(accu);
+    return NULL;
+}
+
+PyDoc_STRVAR(map_doc,
+"map(function, sequence) -> list\n\
+\n\
+Return a list of the results of applying the function to the items of\n\
+the argument sequence.");
+
+/* functional.filter ********************************************************/
+
+static PyObject *
+functional_filter(PyObject *self, PyObject *args)
+{
+    PyObject *seq, *func, *accu, *it, *item;
+    int status;
+
+    if (!PyArg_UnpackTuple(args, "filter", 2, 2, &func, &seq))
+		return NULL;
+    
+    if (!PyCallable_Check(func)) {
+		PyErr_SetString(PyExc_TypeError,
+				"filter() argument 1 must be callable");
+        return NULL;
+	}
+    
+    it = PyObject_GetIter(seq);
+    if (NULL == it) {
+        PyErr_SetString(PyExc_TypeError,
+				"filter() argument 2 must be iterable");
+        return NULL;
+    }
+    
+    accu = PyList_New(0);
+    if (NULL == accu) {
+        Py_DECREF(it);
+        return NULL;
+    }
+    
+    if ((PyObject *)&PyBool_Type == func) {
+        while ((item = PyIter_Next(it))) {
+            if (PyObject_IsTrue(item)) {
+                status = PyList_Append(accu, item);
+                Py_DECREF(item);
+                if (status < 0)
+                    goto fail;
+            }
+            else {
+                Py_DECREF(item);
+            }
+        }
+        if (PyErr_Occurred())
+            goto fail;
+    }
+    else {
+        PyObject *arg, *result;
+
+        arg = PyTuple_New(1);
+        if (NULL == arg)
+            goto fail;
+        
+        while ((item = PyIter_Next(it))) {
+            PyTuple_SET_ITEM(arg, 0, item);
+            result = PyObject_CallObject(func, arg);
+            PyTuple_SET_ITEM(arg, 0, NULL);
+            
+            if (NULL == result) {
+                Py_DECREF(item);
+                Py_DECREF(arg);
+                goto fail;
+            }
+            
+            if (PyObject_IsTrue(result)) {
+                Py_DECREF(result);
+                
+                status = PyList_Append(accu, item);
+                Py_DECREF(item);
+                if (status < 0) {
+                    Py_DECREF(arg);
+                    goto fail;
+                }
+            }
+            else {
+                Py_DECREF(item);
+                Py_DECREF(result);
+            }
+        }
+        Py_DECREF(arg);
+        
+        if (PyErr_Occurred())
+            goto fail;
+    }
+    
+    Py_DECREF(it);
+    return accu;
+
+fail:
+    Py_DECREF(it);
+    Py_DECREF(accu);
+    return NULL;
+}
+
+PyDoc_STRVAR(filter_doc,
+"filter(function, sequence) -> list\n\
+\n\
+Return a list ofthose items in sequence for which function(item) is true");
+
 /* module level code ********************************************************/
 
 PyDoc_STRVAR(module_doc,
 "Tools for functional programming.");
 
-static PyMethodDef module_methods[] = {
+static PyMethodDef functional_methods[] = {
 	{"foldl",	functional_foldl,   METH_VARARGS, foldl_doc},
 	{"foldr",	functional_foldr,   METH_VARARGS, foldr_doc},
 	{"id",      functional_id,      METH_O,       id_doc},
 	{"scanl",   functional_scanl,   METH_VARARGS, scanl_doc},
 	{"scanr",   functional_scanr,   METH_VARARGS, scanr_doc},
+    {"map",     functional_map,     METH_VARARGS, map_doc},
+    {"filter",  functional_filter,  METH_VARARGS, filter_doc},
  	{NULL,		NULL}		/* sentinel */
 };
 
@@ -859,7 +1102,7 @@ initfunctional(void)
 		NULL
 	};
 
-	m = Py_InitModule3("functional", module_methods, module_doc);
+	m = Py_InitModule3("functional", functional_methods, module_doc);
 	if (m == NULL)
 		return;
 
